import numpy as np
import pandas as pd
import scanpy as sc
from scipy.sparse.csgraph import connected_components

from ..utils import check_batch
from .utils import RootCellError


def trajectory_conservation(
    adata_pre, adata_post, label_key, pseudotime_key="dpt_pseudotime", batch_key=None
):
    """Trajectory conservation score

    Trajectory conservation is measured by  spearman’s rank correlation coefficient :math:`s`, between the pseudotime
    values before and after integration.
    The final score was scaled to a value between 0 and 1 using the equation

     .. math::

        trajectory \\, conservation = \\frac {s + 1} {2}

    :param adata_pre: unintegrated adata
    :param adata_post: integrated adata
    :param label_key: column in ``adata_pre.obs`` of the groups used to precompute the trajectory
    :param pseudotime_key: column in ``adata_pre.obs`` in which the pseudotime is saved in.
        Column can contain empty entries, the dataset will be subset to the cells with scores.
    :param batch_key: set to batch key if you want to compute the trajectory metric by batch. By default the batch
        information will be ignored (``batch_key=None``)

    This function requires pseudotime values in ``.obs`` of the unintegrated object (``adata_pre``) computed per batch
    and can be applied to all integration output types.
    The input trajectories should be curated manually as the quality of the metric depends on the quality of the metric
    depends on the quality of the annotation.
    The integrated object (``adata_post``) needs to have a kNN graph based on the integration output.
    See :ref:`preprocessing` for more information on preprocessing.

    **Examples**

    .. code-block:: python

        # feature output
        scib.pp.reduce_data(
            adata, n_top_genes=2000, batch_key="batch", pca=True, neighbors=True
        )
        scib.me.trajectory_conservation(adata_unintegrated, adata, label_key="cell_type")

        # embedding output
        sc.pp.neighbors(adata, use_rep="X_emb")
        scib.me.trajectory_conservation(adata_unintegrated, adata, label_key="celltype")

        # knn output
        scib.me.trajectory_conservation(adata_unintegrated, adata, label_key="celltype")

    """
    # subset to cells for which pseudotime has been computed
    cell_subset = adata_pre.obs.index[adata_pre.obs[pseudotime_key].notnull()]
    adata_pre_ti = adata_pre[cell_subset]
    adata_post_ti = adata_post[cell_subset]
    try:
        iroot, adata_post_ti2 = get_root(
            adata_pre_ti, adata_post_ti, label_key, pseudotime_key
        )
    except RootCellError:
        print("No root cell found, setting trajectory conservation metric to 0.")
        return 0  # failure to find root cell means no TI conservation

    adata_post_ti2.uns["iroot"] = iroot

    sc.tl.dpt(adata_post_ti2)  # stored in 'dpt_pseudotime'
    adata_post_ti2.obs.loc[
        adata_post_ti2.obs["dpt_pseudotime"] > 1, "dpt_pseudotime"
    ] = 0
    adata_post_ti.obs["dpt_pseudotime"] = 0
    adata_post_ti.obs["dpt_pseudotime"] = adata_post_ti2.obs["dpt_pseudotime"]
    adata_post_ti.obs["dpt_pseudotime"].fillna(0, inplace=True)

    if batch_key is None:
        pseudotime_before = adata_pre_ti.obs[pseudotime_key]
        pseudotime_after = adata_post_ti.obs["dpt_pseudotime"]
        correlation = pseudotime_before.corr(pseudotime_after, "spearman")
        return (correlation + 1) / 2  # scaled
    else:
        check_batch(batch_key, adata_pre.obs)
        check_batch(batch_key, adata_post.obs)

        # check if batches match
        if not np.array_equal(
            adata_post_ti.obs[batch_key], adata_pre_ti.obs[batch_key]
        ):
            raise ValueError(
                "Batch columns do not match\n"
                f"adata_post_ti.obs['batch']:\n {adata_post_ti.obs[batch_key]}\n"
                f"adata_pre_ti.obs['batch']:\n {adata_pre_ti.obs[batch_key]}\n"
            )

        corr = pd.Series()
        for i in adata_pre_ti.obs[batch_key].unique():
            pseudotime_before = adata_pre_ti.obs[adata_pre_ti.obs[batch_key] == i][
                pseudotime_key
            ]
            pseudotime_after = adata_post_ti.obs[adata_post_ti.obs[batch_key] == i][
                "dpt_pseudotime"
            ]
            corr[i] = pseudotime_before.corr(pseudotime_after, "spearman")

        return (corr.mean() + 1) / 2  # scaled


def get_root(adata_pre, adata_post, ct_key, pseudotime_key="dpt_pseudotime", dpt_dim=3):
    """Determine root cell for integrated adata based on unintegrated adata

    :param adata_pre: unintegrated adata
    :param adata_post: integrated adata
    :param label_key: column in ``adata_pre.obs`` of the groups used to precompute the trajectory
    :param pseudotime_key: column in ``adata_pre.obs`` in which the pseudotime is saved in.
        Column can contain empty entries, the dataset will be subset to the cells with scores.
    :param dpt_dim: number of diffmap dimensions used to determine root
    """
    n_components, adata_post.obs["neighborhood"] = connected_components(
        csgraph=adata_post.obsp["connectivities"], directed=False, return_labels=True
    )

    start_clust = adata_pre.obs.groupby([ct_key]).mean()[pseudotime_key].idxmin()
    min_dpt = adata_pre.obs[adata_pre.obs[ct_key] == start_clust].index
    which_max_neigh = (
        adata_post.obs["neighborhood"]
        == adata_post.obs["neighborhood"].value_counts().idxmax()
    )
    min_dpt = [
        value for value in min_dpt if value in adata_post.obs[which_max_neigh].index
    ]

    adata_post_ti = adata_post[which_max_neigh]

    min_dpt = [adata_post_ti.obs_names.get_loc(i) for i in min_dpt]

    # compute Diffmap for adata_post
    sc.tl.diffmap(adata_post_ti)

    # determine most extreme cell in adata_post Diffmap
    min_dpt_cell = np.zeros(len(min_dpt))
    for dim in np.arange(dpt_dim):

        diffmap_mean = adata_post_ti.obsm["X_diffmap"][:, dim].mean()
        diffmap_min_dpt = adata_post_ti.obsm["X_diffmap"][min_dpt, dim]

        # count opt cell
        if len(diffmap_min_dpt) == 0:
            raise RootCellError("No root cell in largest component")

        # choose optimum function
        if len(diffmap_min_dpt) > 0 and diffmap_min_dpt.mean() < diffmap_mean:
            opt = np.argmin
        else:
            opt = np.argmax

        min_dpt_cell[opt(diffmap_min_dpt)] += 1

    # root cell is cell with max vote
    return min_dpt[np.argmax(min_dpt_cell)], adata_post_ti



import numpy as np
import pandas as pd

from ..preprocessing import score_cell_cycle
from ..utils import check_adata
from .pcr import pc_regression


def cell_cycle(
    adata_pre,
    adata_post,
    batch_key,
    embed=None,
    agg_func=np.mean,
    organism="mouse",
    n_comps=50,
    recompute_cc=True,
    precompute_pcr_key=None,
    verbose=False,
):
    """Cell cycle conservation score

    Compare the variance contribution of S-phase and G2/M-phase cell cycle scores before and
    after integration. Cell cycle scores are computed per batch on the unintegrated data set,
    eliminating the batch effect confounded by the ``batch_key`` variable.

    .. math::

        CC \\, conservation = 1 - \\frac { |Var_{after} - Var_{before}| } {Var_{before}}

    Variance contribution is obtained through principal component regression using :func:`~scib.metrics.pc_regression`.

    :param adata_pre: adata before integration
    :param adata_post: adata after integration
    :param batch_key: Batch key in ``adata_post.obs``
    :param embed: Name of embedding in ``adata_post.obsm``.
        If ``embed=None``, use the full expression matrix (``adata_post.X``), otherwise use the
        embedding provided in ``adata_post.obsm[embed]``
    :param agg_func: any function that takes a list of numbers and aggregates them into
        a single value. If ``agg_func=None``, all results will be returned
    :param organism: 'mouse' or 'human' for choosing cell cycle genes
    :param n_comps: number of principle components
    :param recompute_cc: If True, force recompute cell cycle score, otherwise use
        precomputed scores if available as 'S_score' and 'G2M_score' in ``adata_post.obs``
    :param precompute_pcr_key: Key in adata_pre for precomputed PCR values for cell
        cycle scores. Ignores cell cycle scores in adata_pre if present.

    :return:
        A score between 1 and 0. The larger the score, the stronger the cell cycle
        variance is conserved.

    The function can be computed on full corrected feature spaces and latent embeddings for both integrated and
    unintegrated ``anndata.Anndata`` objects.
    No preprocessing is needed, as the function will perform PCA directly on the feature or embedding space.

    **Examples**

    .. code-block:: python

        # full feature output
        scib.me.cell_cycle(adata_unintegrated, adata, batch_key="batch")

        # embedding output
        scib.me.cell_cycle(adata_unintegrated, adata, batch_key="batch", embed="X_emb")

    """
    check_adata(adata_pre)
    check_adata(adata_post)

    if embed == "X_pca":
        embed = None

    batches = adata_pre.obs[batch_key].unique()
    scores_final = []
    scores_before = []
    scores_after = []

    recompute_cc = (
        recompute_cc
        or "S_score" not in adata_pre.obs_keys()
        or "G2M_score" not in adata_pre.obs_keys()
    )
    recompute_pcr = (
        precompute_pcr_key is None or precompute_pcr_key not in adata_pre.uns_keys()
    )

    for batch in batches:
        before, after = get_pcr_before_after(
            adata_pre,
            adata_post,
            batch_key=batch_key,
            batch=batch,
            embed=embed,
            organism=organism,
            recompute_cc=recompute_cc,
            recompute_pcr=recompute_pcr,
            pcr_key=precompute_pcr_key,
            n_comps=n_comps,
            verbose=verbose,
        )

        # scale result
        score = 1 - abs(after - before) / before

        if score < 0:
            # Here variance contribution becomes more than twice as large as before
            if verbose:
                print(
                    "Variance contrib more than twice as large after integration.\n"
                    "Setting cell cycle score to 0."
                )
            score = 0

        if verbose:
            print(
                f"batch: {batch}\t before: {before}\t after: {after}\t score: {score}"
            )

        scores_before.append(before)
        scores_after.append(after)
        scores_final.append(score)

    if agg_func is None:
        return pd.DataFrame(
            [batches, scores_before, scores_after, scores_final],
            columns=["batch", "before", "after", "score"],
        )
    else:
        return agg_func(scores_final)


def get_pcr_before_after(
    adata_pre,
    adata_post,
    batch_key,
    batch,
    embed,
    organism,
    recompute_cc,
    recompute_pcr,
    pcr_key,
    n_comps,
    verbose,
):
    """
    Principle component regression value on cell cycle scores for one batch
    before and after integration

    :param adata_pre: adata before integration
    :param adata_post: adata after integration
    :param embed: Name of embedding in adata_post.obsm.
        If ``embed=None``, use the full expression matrix (``adata.X``), otherwise use the
        embedding provided in ``adata_post.obsm[embed]``
    :param organism: 'mouse' or 'human' for choosing cell cycle genes
    :param recompute_cc: If True, force recompute cell cycle score, otherwise use
        precomputed scores if available as 'S_score' and 'G2M_score' in adata.obs
    :param recompute_pcr: Whether to recompute principal component regression.
    :param pcr_key: Key in adata_pre for precomputed PCR values for cell cycle scores.
    :param n_comps: Number of components for PCA
    :param verbose:

    :return:
        A score between 1 and 0. The larger the score, the stronger the cell cycle
        variance is conserved.
    """
    # subset adata objects
    raw_sub = adata_pre[adata_pre.obs[batch_key] == batch]
    int_sub = adata_post[adata_post.obs[batch_key] == batch].copy()
    int_sub = int_sub.obsm[embed] if embed is not None else int_sub.X

    # sanity check: subsets have same number of rows?
    if raw_sub.shape[0] != int_sub.shape[0]:
        raise ValueError(
            f'batch "{batch}" of batch_key "{batch_key}" has unequal number of '
            f"entries before and after integration.\n"
            f"before: {raw_sub.shape[0]} after: {int_sub.shape[0]}"
        )

    # compute cc scores if necessary
    if recompute_cc:
        if verbose:
            print("Score cell cycle...")
        score_cell_cycle(raw_sub, organism=organism)

    # regression variable
    covariate = raw_sub.obs[["S_score", "G2M_score"]]

    # PCR on adata before integration
    if recompute_pcr:
        before = pc_regression(
            raw_sub.X, covariate, pca_var=None, n_comps=n_comps, verbose=verbose
        )
    else:
        before = pd.Series(raw_sub.uns[pcr_key])

    # PCR on adata after integration
    after = pc_regression(
        int_sub, covariate, pca_var=None, n_comps=n_comps, verbose=verbose
    )

    return before, after



import logging

import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse

from ..exceptions import OptionalDependencyNotInstalled, RLibraryNotFound
from ..utils import check_adata, check_batch
from .utils import NeighborsError, diffusion_conn, diffusion_nn


def kBET(
    adata,
    batch_key,
    label_key,
    type_,
    embed=None,
    scaled=True,
    return_df=False,
    verbose=False,
):
    """kBET score

    Compute the average of k-nearest neighbour batch effect test (kBET) score per label.
    This is a wrapper function of the implementation by `Büttner et al. 2019`_.
    kBET measures the bias of a batch variable in the kNN graph.
    Specifically, kBET is quantified as the average rejection rate of Chi-squared tests of local vs  global batch label
    distributions.
    This means that smaller values indicate better batch mixing.
    By default the original kBET score is scaled between 0 and 1 so that larger scores are associated with better batch
    mixing.

    .. _Büttner et al. 2019: https://doi.org/10.1038/s41592-018-0254-1

    :param adata: anndata object to compute kBET on
    :param batch_key: name of batch column in adata.obs
    :param label_key: name of cell identity labels column in adata.obs
    :param `type_`: type of data integration, one of 'knn', 'embed' or 'full'
    :param embed: embedding key in ``adata.obsm`` for embedding and feature input
    :param scaled: whether to scale between 0 and 1
        with 0 meaning low batch mixing and 1 meaning optimal batch mixing
        if scaled=False, 0 means optimal batch mixing and 1 means low batch mixing
    :return:
        kBET score (average of kBET per label) based on observed rejection rate.
        If ``return_df=True``, also return a ``pd.DataFrame`` with kBET observed
        rejection rate per cluster

    This function can be applied to all integration output types and recomputes the kNN graph for feature and embedding
    output with specific parameters.
    Thus, no preprocessing is required, but the correct output type must be specified in ``type_``.

    **Examples**

    .. code-block:: python

        # full feature integration output or unintegrated data
        scib.me.kBET(
            adata, batch_key="batch", label_key="celltype", type_="full", embed="X_pca"
        )

        # embedding output
        scib.me.kBET(
            adata, batch_key="batch", label_key="celltype", type_="embed", embed="X_emb"
        )

        # kNN output
        scib.me.kBET(adata, batch_key="batch", label_key="celltype", type_="knn")

    """
    try:
        import rpy2.rinterface_lib.callbacks
        import rpy2.rinterface_lib.embedded
        import rpy2.robjects as ro

        rpy2.rinterface_lib.callbacks.logger.setLevel(logging.ERROR)
    except ModuleNotFoundError as e:
        raise OptionalDependencyNotInstalled(e)

    check_adata(adata)
    check_batch(batch_key, adata.obs)
    check_batch(label_key, adata.obs)

    try:
        ro.r("library(kBET)")
    except rpy2.rinterface_lib.embedded.RRuntimeError as e:
        raise OptionalDependencyNotInstalled(e, "kBET")

    # compute connectivities for non-knn type data integrations
    # and increase neighborhoods for knn type data integrations
    if type_ != "knn" and embed is not None:
        adata_tmp = sc.pp.neighbors(adata, n_neighbors=50, use_rep=embed, copy=True)
    else:
        # check if pre-computed neighbours are stored in input file
        adata_tmp = adata.copy()
        if "diffusion_connectivities" not in adata.uns["neighbors"]:
            if verbose:
                print("Compute diffusion neighbours")
            adata_tmp = diffusion_conn(adata, min_k=50, copy=True)
        adata_tmp.obsp["connectivities"] = adata_tmp.uns["neighbors"][
            "diffusion_connectivities"
        ]

    if verbose:
        print(f"batch: {batch_key}")

    # set upper bound for k0
    size_max = 2**31 - 1

    # prepare call of kBET per cluster
    kBET_scores = {"cluster": [], "kBET": []}
    for clus in adata_tmp.obs[label_key].unique():

        # subset by label
        adata_sub = adata_tmp[adata_tmp.obs[label_key] == clus, :].copy()

        # check if neighborhood size too small or only one batch in subset
        if np.logical_or(
            adata_sub.n_obs < 10, len(adata_sub.obs[batch_key].cat.categories) == 1
        ):
            print(f"{clus} consists of a single batch or is too small. Skip.")
            score = np.nan
        else:
            quarter_mean = np.floor(
                np.mean(adata_sub.obs[batch_key].value_counts()) / 4
            ).astype("int")
            k0 = np.min([70, np.max([10, quarter_mean])])
            # check k0 for reasonability
            if k0 * adata_sub.n_obs >= size_max:
                k0 = np.floor(size_max / adata_sub.n_obs).astype("int")

            matrix = np.zeros(shape=(adata_sub.n_obs, k0 + 1))

            if verbose:
                print(f"Use {k0} nearest neighbors.")
            n_comp, labs = scipy.sparse.csgraph.connected_components(
                adata_sub.obsp["connectivities"], connection="strong"
            )

            if n_comp == 1:  # a single component to compute kBET on
                try:
                    nn_index_tmp = diffusion_nn(adata_sub, k=k0).astype("float")
                    # call kBET
                    score = kBET_single(
                        matrix=matrix,
                        batch=adata_sub.obs[batch_key],
                        knn=nn_index_tmp
                        + 1,  # nn_index in python is 0-based and 1-based in R
                        verbose=verbose,
                        k0=k0,
                    )
                except NeighborsError:
                    print("Not enough neighbours")
                    score = 1  # i.e. 100% rejection

            else:
                # check the number of components where kBET can be computed upon
                comp_size = pd.value_counts(labs)
                # check which components are small
                comp_size_thresh = 3 * k0
                idx_nonan = np.flatnonzero(
                    np.in1d(labs, comp_size[comp_size >= comp_size_thresh].index)
                )

                # check if 75% of all cells can be used for kBET run
                if len(idx_nonan) / len(labs) >= 0.75:
                    # create another subset of components, assume they are not visited in a diffusion process
                    adata_sub_sub = adata_sub[idx_nonan, :].copy()
                    nn_index_tmp = np.empty(shape=(adata_sub.n_obs, k0))
                    nn_index_tmp[:] = np.nan

                    try:
                        nn_index_tmp[idx_nonan] = diffusion_nn(
                            adata_sub_sub, k=k0
                        ).astype("float")
                        # call kBET
                        score = kBET_single(
                            matrix=matrix,
                            batch=adata_sub.obs[batch_key],
                            knn=nn_index_tmp
                            + 1,  # nn_index in python is 0-based and 1-based in R
                            verbose=verbose,
                            k0=k0,
                        )
                    except NeighborsError:
                        print("Not enough neighbours")
                        score = 1  # i.e. 100% rejection
                else:  # if there are too many too small connected components, set kBET score to 1
                    score = 1  # i.e. 100% rejection

        kBET_scores["cluster"].append(clus)
        kBET_scores["kBET"].append(score)

    kBET_scores = pd.DataFrame.from_dict(kBET_scores)
    kBET_scores = kBET_scores.reset_index(drop=True)

    if return_df:
        return kBET_scores

    final_score = np.nanmean(kBET_scores["kBET"])
    return 1 - final_score if scaled else final_score


def kBET_single(matrix, batch, k0=10, knn=None, verbose=False):
    """Single kBET run

    Compute k-nearest neighbour batch effect test (kBET) score as described in
    https://doi.org/10.1038/s41592-018-0254-1

    :param matrix: expression matrix (at the moment: a PCA matrix, so ``do.pca`` is set to ``FALSE``)
    :param batch: series or list of batch assignments
    :returns: kBET observed rejection rate
    """
    try:
        import anndata2ri
        import rpy2.rinterface_lib.callbacks
        import rpy2.rinterface_lib.embedded
        import rpy2.robjects as ro

        rpy2.rinterface_lib.callbacks.logger.setLevel(logging.ERROR)
    except ModuleNotFoundError as e:
        raise OptionalDependencyNotInstalled(e)

    try:
        ro.r("library(kBET)")
    except rpy2.rinterface_lib.embedded.RRuntimeError as ex:
        RLibraryNotFound(ex)

    anndata2ri.activate()

    if verbose:
        print("importing expression matrix")
    ro.globalenv["data_mtrx"] = matrix
    ro.globalenv["batch"] = batch

    if verbose:
        print("kBET estimation")

    ro.globalenv["knn_graph"] = knn
    ro.globalenv["k0"] = k0
    ro.r(
        "batch.estimate <- kBET("
        "  data_mtrx,"
        "  batch,"
        "  knn=knn_graph,"
        "  k0=k0,"
        "  plot=FALSE,"
        "  do.pca=FALSE,"
        "  heuristic=FALSE,"
        "  adapt=FALSE,"
        f"  verbose={str(verbose).upper()}"
        ")"
    )

    try:
        score = ro.r("batch.estimate$summary$kBET.observed")[0]
    except rpy2.rinterface_lib.embedded.RRuntimeError as ex:
        print(f"Error computing kBET: {ex}\nSetting value to np.nan")
        score = np.nan

    anndata2ri.deactivate()

    return score



import os
import subprocess

from scanpy._utils import deprecated_arg_names
from sklearn.metrics.cluster import normalized_mutual_info_score

from ..utils import check_adata, check_batch


@deprecated_arg_names(
    {"group1": "cluster_key", "group2": "label_key", "method": "implementation"}
)
def nmi(adata, cluster_key, label_key, implementation="arithmetic", nmi_dir=None):
    """Normalized mutual information

    The normalized mutual information is a version of the mutual information corrected by the entropy of clustering and
    ground truth labels (e.g. cell type).
    The score ranges between 0 and 1, with 0 representing no sharing and 1 representing perfect sharing of information
    between clustering and annotated cell labels.

    :param adata: anndata object with cluster assignments in ``adata.obs[cluster_key]``
    :param cluster_key: string of column in adata.obs containing cluster assignments
    :param label_key: string of column in adata.obs containing labels
    :param implementation: NMI implementation.
        ``'max'``: scikit method with ``average_method='max'``;
        ``'min'``: scikit method with ``average_method='min'``;
        ``'geometric'``: scikit method with ``average_method='geometric'``;
        ``'arithmetic'``: scikit method with ``average_method='arithmetic'``;
        ``'Lancichinetti'``: implementation by A. Lancichinetti 2009 et al. https://sites.google.com/site/andrealancichinetti/mutual;
        ``'ONMI'``: implementation by Aaron F. McDaid et al. https://github.com/aaronmcdaid/Overlapping-NMI
    :param nmi_dir: directory of compiled C code if 'Lancichinetti' or 'ONMI' are specified as ``method``.
        These packages need to be compiled as specified in the corresponding READMEs.
    :return: Normalized mutual information NMI value

    This function can be applied to all integration output types.
    The ``adata`` must contain cluster assignments that are based off the knn graph given or derived from the integration
    method output.
    For this metric you need to include all steps that are needed for clustering.
    See :ref:`preprocessing` for more information on preprocessing.

    **Examples**

    .. code-block:: python

        # feature output
        scib.pp.reduce_data(
            adata, n_top_genes=2000, batch_key="batch", pca=True, neighbors=True
        )
        scib.me.cluster_optimal_resolution(adata, cluster_key="cluster", label_key="celltype")
        scib.me.nmi(adata, cluster_key="cluster", label_key="celltype")

        # embedding output
        sc.pp.neighbors(adata, use_rep="X_emb")
        scib.me.cluster_optimal_resolution(adata, cluster_key="cluster", label_key="celltype")
        scib.me.nmi(adata, cluster_key="cluster", label_key="celltype")

        # knn output
        scib.me.cluster_optimal_resolution(adata, cluster_key="cluster", label_key="celltype")
        scib.me.nmi(adata, cluster_key="cluster", label_key="celltype")

    """

    check_adata(adata)
    check_batch(cluster_key, adata.obs)
    check_batch(label_key, adata.obs)

    cluster_key = adata.obs[cluster_key].tolist()
    label_key = adata.obs[label_key].tolist()

    if len(cluster_key) != len(label_key):
        raise ValueError(
            f"different lengths in cluster_key ({len(cluster_key)}) and label_key ({len(label_key)})"
        )

    # choose method
    if implementation in ["max", "min", "geometric", "arithmetic"]:
        nmi_value = normalized_mutual_info_score(
            cluster_key, label_key, average_method=implementation
        )
    elif implementation == "Lancichinetti":
        nmi_value = nmi_Lanc(cluster_key, label_key, nmi_dir=nmi_dir)
    elif implementation == "ONMI":
        nmi_value = onmi(cluster_key, label_key, nmi_dir=nmi_dir)
    else:
        raise ValueError(f"Method {implementation} not valid")

    return nmi_value


def onmi(group1, group2, nmi_dir=None, verbose=True):
    """
    Based on implementation https://github.com/aaronmcdaid/Overlapping-NMI
    publication: Aaron F. McDaid, Derek Greene, Neil Hurley 2011

    :param group1: list or series of cell assignments
    :param group2: list or series of cell assignments
    :param nmi_dir: directory of compiled C code
    """

    if nmi_dir is None:
        raise FileNotFoundError(
            "Please provide the directory of the compiled C code from "
            "https://sites.google.com/site/andrealancichinetti/mutual3.tar.gz"
        )

    group1_file = write_tmp_labels(group1, to_int=False)
    group2_file = write_tmp_labels(group2, to_int=False)

    nmi_call = subprocess.Popen(
        [nmi_dir + "onmi", group1_file, group2_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    stdout, stderr = nmi_call.communicate()
    if stderr:
        print(stderr)

    nmi_out = stdout.decode()
    if verbose:
        print(nmi_out)

    nmi_split = [x.strip().split("\t") for x in nmi_out.split("\n")]
    nmi_max = float(nmi_split[0][1])

    # remove temporary files
    os.remove(group1_file)
    os.remove(group2_file)

    return nmi_max


def nmi_Lanc(group1, group2, nmi_dir="external/mutual3/", verbose=True):
    """
    paper by A. Lancichinetti 2009
    https://sites.google.com/site/andrealancichinetti/mutual
    recommended by Malte
    """

    if nmi_dir is None:
        raise FileNotFoundError(
            "Please provide the directory of the compiled C code from https://sites.google.com/site/andrealancichinetti/mutual3.tar.gz"
        )

    group1_file = write_tmp_labels(group1, to_int=False)
    group2_file = write_tmp_labels(group2, to_int=False)

    nmi_call = subprocess.Popen(
        [nmi_dir + "mutual", group1_file, group2_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    stdout, stderr = nmi_call.communicate()
    if stderr:
        print(stderr)
    nmi_out = stdout.decode().strip()

    return float(nmi_out.split("\t")[1])


def write_tmp_labels(group_assignments, to_int=False, delim="\n"):
    """
    write the values of a specific obs column into a temporary file in text format
    needed for external C NMI implementations (onmi and nmi_Lanc functions), because they require files as input

    :param group_assignments: list or series of cell assignments
    :param to_int: rename the unique column entries by integers in range(1,len(group_assignments)+1)
    """
    import tempfile

    if to_int:
        label_map = {}
        i = 1
        for label in set(group_assignments):
            label_map[label] = i
            i += 1
        delim.join([str(label_map[name]) for name in group_assignments])
    else:
        delim.join([str(name) for name in group_assignments])

    clusters = {label: [] for label in set(group_assignments)}
    for i, label in enumerate(group_assignments):
        clusters[label].append(str(i))

    output = "\n".join([" ".join(c) for c in clusters.values()])
    output = str.encode(output)

    # write to file
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(output)
        filename = f.name

    return filename



import numpy as np
import pandas as pd
import scipy.special
from scanpy._utils import deprecated_arg_names
from sklearn.metrics.cluster import adjusted_rand_score

from ..utils import check_adata, check_batch


@deprecated_arg_names({"group1": "cluster_key", "group2": "label_key"})
def ari(adata, cluster_key, label_key, implementation=None):
    """Adjusted Rand Index

    The adjusted rand index is a chance-adjusted rand index, which evaluates the pair-wise accuracy of clustering vs.
    ground truth label assignments.
    The score ranges between 0 and 1 with larger values indicating better conservation of the data-driven cell identity
    discovery after integration compared to annotated labels.

    :param adata: anndata object with cluster assignments in ``adata.obs[cluster_key]``
    :param cluster_key: string of column in adata.obs containing cluster assignments
    :param label_key: string of column in adata.obs containing labels
    :param implementation: if set to 'sklearn', uses sklearn's implementation,
        otherwise native implementation is taken

    This function can be applied to all integration output types.
    The ``adata`` must contain cluster assignments that are based off the knn graph given or derived from the integration
    method output.
    For this metric you need to include all steps that are needed for clustering.
    See :ref:`preprocessing` for more information on preprocessing.

    **Examples**

    .. code-block:: python

        # feature output
        scib.pp.reduce_data(
            adata, n_top_genes=2000, batch_key="batch", pca=True, neighbors=True
        )
        scib.me.cluster_optimal_resolution(adata, cluster_key="cluster", label_key="celltype")
        scib.me.ari(adata, cluster_key="cluster", label_key="celltype")

        # embedding output
        sc.pp.neighbors(adata, use_rep="X_emb")
        scib.me.cluster_optimal_resolution(adata, cluster_key="cluster", label_key="celltype")
        scib.me.ari(adata, cluster_key="cluster", label_key="celltype")

        # knn output
        scib.me.cluster_optimal_resolution(adata, cluster_key="cluster", label_key="celltype")
        scib.me.ari(adata, cluster_key="cluster", label_key="celltype")

    """

    check_adata(adata)
    check_batch(cluster_key, adata.obs)
    check_batch(label_key, adata.obs)

    cluster_key = adata.obs[cluster_key].to_numpy()
    label_key = adata.obs[label_key].to_numpy()

    if len(cluster_key) != len(label_key):
        raise ValueError(
            f"different lengths in cluster_key ({len(cluster_key)}) and label_key ({len(label_key)})"
        )

    if implementation == "sklearn":
        return adjusted_rand_score(cluster_key, label_key)

    def binom_sum(x, k=2):
        return scipy.special.binom(x, k).sum()

    n = len(cluster_key)
    contingency = pd.crosstab(cluster_key, label_key)

    ai_sum = binom_sum(contingency.sum(axis=0))
    bi_sum = binom_sum(contingency.sum(axis=1))

    index = binom_sum(np.ravel(contingency))
    expected_index = ai_sum * bi_sum / binom_sum(n, 2)
    max_index = 0.5 * (ai_sum + bi_sum)

    return (index - expected_index) / (max_index - expected_index)
